/*
  MethOS kernel initialization and shell startup
 */

#include <oskit/clientos.h>
#include <oskit/debug.h>
#include <oskit/exec/exec.h>
#include <oskit/gdb.h>
#include <oskit/machine/pc/phys_lmm.h>
#include <oskit/sproc.h>
#include <oskit/startup.h>
#include <oskit/threads/pthread.h>
#include <oskit/version.h>
#include <oskit/x86/proc_reg.h>
#include <oskit/x86/trap.h>
#include <oskit/page.h>
#include <oskit/dev/osenv.h>
#include <oskit/dev/freebsd.h>
#include <oskit/c/termios.h>
#include <oskit/c/unistd.h>
#include <oskit/c/stdio.h>
#include <oskit/c/stdlib.h>
#include <oskit/c/malloc.h>
#include <oskit/c/fcntl.h>

#include "kernel/process.h"
#include "kernel/message.h"

extern int
main()
{
	int shellprocess = 0;
	int success = 0;
	oskit_osenv_t *osenv = 0;

	oskit_clientos_init_pthreads();
	osenv = start_osenv ();
	start_clock();
	start_pthreads ();
//	oskit_dev_init (osenv);
//	oskit_freebsd_init_osenv (osenv);
//	oskit_freebsd_init_sio ();
//	osenv_process_lock ();
	start_devices ();
//	osenv_process_unlock ();
	start_fs_bmod();
//	start_console_pthreads ();

	oskit_uvm_init();
	oskit_uvm_swap_init();

	if (swapon("/swapfile")) {
		extern int errno;
		panic("swapon failed (errno %d)\n", errno);
	}

	oskit_uvm_start_pagedaemon();

	oskit_sproc_init();
	
	printf ("\nMethOS\n\n");

	/* Initialize my system */
	mth_syscall_init ();
	mth_process_init ();
	mth_message_init ();

	shellprocess = mth_process_create ();  

	success = mth_process_loadelf (shellprocess, "kest");
	assert (success == 0);

	mth_process_join (shellprocess);

	printf ("\nNyyd v6id arvuti sulgeda.\n\n");
	
	return 0;
}

void
dump_kvmspace()
{
	oskit_uvm_vmspace_dump(&oskit_uvm_kvmspace);
}

