/*
  MethOS syscall implementations
*/

#include <oskit/machine/types.h>
#include <oskit/threads/pthread.h>
#include <oskit/c/sys/stat.h>
#include <oskit/c/sys/types.h>
#include <oskit/c/sys/time.h>
#include <oskit/c/unistd.h>
#include <oskit/c/stdio.h>
#include <oskit/c/stdlib.h>
#include <oskit/c/dirent.h>
#include <oskit/c/string.h>
#include <oskit/c/malloc.h>
#include <oskit/c/errno.h>
#include <oskit/c/fcntl.h>
#include <oskit/c/time.h>
#include <oskit/sproc.h>

#include <kernel/process.h>
#include <kernel/message.h>

pthread_mutex_t syslock;

extern void sys_exit(struct oskit_sproc_thread *sth,
		     void *arg, int *rval) OSKIT_NORETURN;

void
mth_syscall_init()
{
	pthread_mutex_init(&syslock, 0);
}

void 
syscall_exit(struct oskit_sproc_thread *sth, void *arg, int *rval)
{
	int exit_code = *(int*)arg;
	OSKIT_SPROC_RETURN(sth, exit_code);
}

int 
syscall_putchar(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
	int c = *(int*)arg;
	*rval = putchar(c);
	return 0;
}

int
syscall_getchar(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
	int a;

	int procnum = 0;
	int type = -1;
	procnum = mth_process_self ();
	while (type != TTYPE_INTERACTIVE) {
		type = mth_process_gettype (procnum);
		if (type == TTYPE_BACKGROUND)
			sched_yield ();
	}

	a = fgetc(stdin);
	*rval = a;
	return 0;
}

int
syscall_printf(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
	const char *fmt = (const char*)(((char**)arg)[0]);
	oskit_va_list va;
	va = (oskit_va_list)(arg + 4);
	*rval = vprintf(fmt, va);
	fflush(stdout);
	return 0;
}

int
syscall_sleep(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
	int t = *(int*)arg;
	sleep(t);
	return 0;
}

int
syscall_puts(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
	struct puts_arg {
		const char *str;
	} *ap;
	char buf[256];
	oskit_size_t sz;
	oskit_error_t error;
	ap = (struct puts_arg*)arg;

	error = copyinstr(ap->str, buf, 256, &sz);
	if (error) {
		return error;
	}
	puts(buf);

	return 0;
}

int
syscall_gets(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
	int procnum = 0;
	int type = -1;
	struct gets_arg {
		char *str;
	} *ap;
	ap = (struct gets_arg*)arg;
	procnum = mth_process_self ();
	while (type != TTYPE_INTERACTIVE) {
		type = mth_process_gettype (procnum);
		if (type == TTYPE_BACKGROUND)
			sched_yield ();
	}
	fgets (ap->str, 256, stdin);
	*rval = ap->str;
	return 0;
}

int
syscall_lock(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
    pthread_mutex_lock(&syslock);
    return 0;
}

int
syscall_unlock(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
    pthread_mutex_unlock(&syslock);
    return 0;
}

int 
syscall_process_create (struct oskit_sproc_thread *proc, void *arg, int *rval) {
	*rval = mth_process_create ();
	return 0;

}

int syscall_process_loadelf (struct oskit_sproc_thread *proc, void *arg, int *rval) {
	struct process_filename_arg {
		int procnum;
		const char* fname;
	} *ap;
	ap = (struct process_filename_arg*)arg;
	*rval = mth_process_loadelf (ap->procnum, ap->fname);
	return 0;    
}

int 
syscall_process_join (struct oskit_sproc_thread *proc, void *arg, int *rval) {
	struct process_arg {
		int procnum;
	} *ap;
	ap = (struct process_arg*)arg;
	mth_process_join (ap->procnum);
	return 0;    
}

int 
syscall_process_kill (struct oskit_sproc_thread *proc, void *arg, int *rval) {
	struct process_arg {
		int procnum;
	} *ap;
	ap = (struct process_arg*)arg;
	mth_process_kill (ap->procnum);
	return 0;    
}

int 
syscall_process_yield (struct oskit_sproc_thread *proc, void *arg, int *rval) {
	struct process_arg {
		int procnum;
	} *ap;
	ap = (struct process_arg*)arg;
	sched_yield ();
	return 0;    
}

int
syscall_process_settype (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct process_settype_arg {
		int procnum;
		int type;
	} *ap;
	ap = (struct process_settype_arg*)arg;
	oskit_pthread_sleep (200);
	mth_process_settype (ap->procnum, ap->type);
	return 0;	
}

int
syscall_process_gettype (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct process_arg {
		int procnum;
	} *ap;
	ap = (struct process_arg*)arg;
	*rval = mth_process_gettype (ap->procnum);
	return 0;	
}

int 
syscall_process_list (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	mth_process_list ();
	return 0;
}

int
syscall_process_number (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct process_number_arg {
		const char *name;
	} *ap;
	ap = (struct process_number_arg*)arg;
	*rval = mth_process_number (ap->name);
	return 0;    	
}

int
syscall_fopen (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct fopen_arg {
		const char* path;
		const char* mode;
	} *ap;
	ap = (struct fopen_arg*)arg;
	FILE* file = fopen (ap->path, ap->mode);
	*rval = file;
	return 0;    
}

int
syscall_fclose (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct fclose_arg {
		FILE* file;
	} *ap;
	ap = (struct fclose_arg*)arg;
	int res = fclose (ap->file);
	*rval = res;
	return 0;    
}

int
syscall_fileexists (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct fopen_arg {
		const char* path;
	} *ap;
	ap = (struct fopen_arg*)arg;
	FILE* file = 0;
	if (ap->path == 0)
		*rval = 1;
	file = fopen (ap->path, "r");
	if (file != 0) {
		*rval = 0;
		fclose (file);
	}
	else {
		*rval = 1;
	}
	return 0;    
}

int
syscall_fprintf(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
	struct fprintf_arg {
		FILE *file;
		char *arg;
	} *ap;
	char a[100];
	ap = (struct fprintf_arg*)arg;
	const char *fmt = (const char*)(((char**)(arg + 4))[0]);
	oskit_va_list va;
	va = (oskit_va_list)(arg + 8);
	vsprintf(a, fmt, va);
//	printf ("fprintf: %s\n", a);
//	fprintf (ap->file, "%s", a);
//	fputs (a, ap->file);
	*rval = vfprintf(ap->file, fmt, va);
	fflush (ap->file);	
	return 0;
}

int
syscall_fputs(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
	struct fputs_arg {
		const char *str;
		FILE* file;
	} *ap;
	char buf[256];
	oskit_size_t sz;
	oskit_error_t error;
	ap = (struct fputs_arg*)arg;

	error = copyinstr(ap->str, buf, 256, &sz);
	if (error) {
		return error;
	}
	fputs(buf, ap->file);

	return 0;
}

int
syscall_fgets(struct oskit_sproc_thread *proc, void *arg, int *rval)
{
	struct fgets_arg {
		char *str;
		int size;
		FILE* file;
	} *ap;
	ap = (struct fgets_arg*)arg;
	fgets (ap->str, ap->size, ap->file);
	*rval = ap->str;
	return 0;
}

int
syscall_feof (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct feof_arg {
		FILE * file;
	} *ap;
	ap = (struct feof_arg*)arg;
	*rval = feof (ap->file);
	return 0;	
}

int
syscall_srand (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct srand_arg {
		unsigned seed;
	} *ap;
	ap = (struct srand_arg*)arg;
	srand (ap->seed);
	return 0;	
}

int
syscall_time (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct time_arg {
		time_t tm;
	} *ap;
	ap = (struct time_arg*)arg;
	*rval = time (ap->tm);
	return 0;	
}

int
syscall_rand (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	*rval = rand ();
	return 0;	
}

int
syscall_atoi (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct atoi_arg {
		const char* npr;
	} *ap;
	ap = (struct atoi_arg*)arg;
	*rval = atoi (ap->npr);
	return 0;	
}
 

int
syscall_message_send (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct message_send_arg {
		int dest;
		int size;
		char *data;
	} *ap;
	ap = (struct message_send_arg*)arg;
	*rval = mth_message_send (ap->dest, ap->size, ap->data);
	return 0;
}

int
syscall_message_receive (struct oskit_sproc_thread *proc, void *arg, int *rval) 
{
	struct message_receive_arg {
		int *sender;
		int *size;
		char *data;
	} *ap;
	ap = (struct message_receive_arg*)arg;
	*rval = mth_message_receive (ap->sender, ap->size, ap->data);
	return 0;    
}

struct oskit_sproc_sysent methos_syscall_table[] = {
	{ (oskit_sproc_syscall_func)syscall_exit, 1 },
	{ syscall_putchar, 1 },
	{ syscall_getchar, 1 },
	{ syscall_printf, 8 },
	{ syscall_sleep, 1 },
	{ syscall_gets, 1 },
	{ syscall_puts, 1 },
	{ syscall_lock, 0 },
	{ syscall_unlock, 0 },
	{ syscall_process_create, 0 },
	{ syscall_process_loadelf, 2 },
	{ syscall_process_join, 1 },	
	{ syscall_process_list, 0 },
	{ syscall_process_number, 1 },
	{ syscall_fopen, 2},
	{ syscall_fclose, 1},
	{ syscall_fileexists, 1},
	{ syscall_message_send, 3},
	{ syscall_message_receive, 3},
	{ syscall_fprintf, 9},
	{ syscall_fgets, 3},
	{ syscall_fputs, 2},
	{ syscall_feof, 1},
	{ syscall_process_kill, 1},
	{ syscall_srand, 1 },
	{ syscall_rand, 0 },
	{ syscall_atoi, 1 },
	{ syscall_time, 1 },
	{ syscall_process_yield, 0},
	{ syscall_process_settype, 2},
	{ syscall_process_gettype, 1}
};
