/*
  MethOS scheduler
  
  Contains portions from OSKit POSIX scheduler
 */

#include <threads/pthread_internal.h>
#include "pthread_signal.h"
#define  COMPILING_SCHEDULER
#include "sched_methos.h"

/* pthread_scheduler.c */
extern pthread_lock_t	pthread_sched_lock;

#define NUMQUEUES 2
#define MIN_TIMESLICE 1
#define MAX_TIMESLICE 4
#define INITIAL_TIMESLICE 2

int queue_count[NUMQUEUES] = {0,0};
int which_process = 0;	// process cycle counter (background on 5 / 10)

/* For accounting */
oskit_u32_t tickcount = 0;
extern oskit_u32_t threads_realticks;

static queue_head_t	threads_runq[NUMQUEUES] = { {0} };
static int		threads_runq_count = 0;
static oskit_u32_t	threads_whichrunq;

extern int		ffs();

pthread_thread_t* idle_thread = 0;

void
methos_sched_setidlethread (int tid)
{
	idle_thread = threads_tidtothread[tid];
}

int
methos_sched_type (int tid)
{
	pthread_thread_t *pthread = threads_tidtothread[tid];
	return pthread->type;
}

void
methos_sched_settype (int tid, int type)
{
	pthread_thread_t *pthread = threads_tidtothread[tid];
	if (type == TTYPE_INTERACTIVE)
		pthread->timeslice = INITIAL_TIMESLICE;
	pthread->type = type;
}

int
methos_sched_timeslice (int tid)
{
	pthread_thread_t *pthread = threads_tidtothread[tid];
	return pthread->timeslice;
}

int
methos_sched_cpupercent (int tid)
{
	pthread_thread_t *pthread = threads_tidtothread[tid];
	return pthread->percent;
}

int
methos_sched_cputime (int tid)
{
	pthread_thread_t *pthread = threads_tidtothread[tid];
	return pthread->cputime;
}

void 
methos_sched_calc_percentages () 
{
	int i = 0;
	pthread_thread_t *thread;
	int sum = 0;

	for (i = 1; i < THREADS_MAX_THREAD; i++) {
		thread = threads_tidtothread[i];
		if (thread != 0) {
      			thread->percent = thread->cputime - thread->totalticks;
			sum += thread->percent;
			thread->totalticks = thread->cputime;
		}
	}

	for (i = 1; i < THREADS_MAX_THREAD; i++) {
		thread = threads_tidtothread[i];
		if (thread != 0) {
      			thread->percent = (thread->percent * 100)/ sum;
		}
	}
}

/*
 *  Thread queue functions (adapted from the posix scheduler)
 */ 

static inline int
methos_runq_empty(void)
{
	return (threads_whichrunq == 0);
}

static inline int
methos_runq_onrunq(pthread_thread_t *pthread)
{
	return (int) pthread->runq.next;
}

static inline void
methos_runq_insert_tail(pthread_thread_t *pthread)
{
	int queue = pthread->type;

	queue_head_t *phdr = &threads_runq[queue];

	queue_count[queue]++;
	queue_enter(phdr, pthread, pthread_thread_t *, runq);
	threads_whichrunq |= (1 << queue);

	threads_runq_count++;
}

static inline void
methos_runq_insert_head(pthread_thread_t *pthread)
{
	int queue = pthread->type;

	queue_head_t *phdr = &threads_runq[queue];

	queue_count[queue]++;

	queue_enter_first(phdr, pthread, pthread_thread_t *, runq);

	threads_whichrunq |= (1 << queue);
	threads_runq_count++;
}


static inline pthread_thread_t *
methos_runq_dequeue(void)
{
	pthread_lock(&pthread_sched_lock);

	queue_head_t *phdr;
	pthread_thread_t *pnext;
	int queue = 0;

	which_process++;
	if (which_process > 10)
		which_process = 1;
	

	if ((which_process == 5 || which_process == 10) && queue_count[TTYPE_BACKGROUND] > 0)
		queue = TTYPE_BACKGROUND;
	else
		queue = TTYPE_INTERACTIVE;

	
	phdr = &threads_runq[queue];

	if (queue_count[queue] < 1) {
		pnext = idle_thread;
	} 
	else {
		queue_remove_first(phdr, pnext, pthread_thread_t *, runq);
		queue_count [queue]--;
    		threads_runq_count--;

		pnext->runq.next = (queue_entry_t) 0;	

		if (queue_empty(phdr))
			threads_whichrunq &= ~(1 << queue);

	}

	pthread_unlock(&pthread_sched_lock);

	return pnext;	
}

static inline void
methos_runq_remove(pthread_thread_t *pthread)
{
	int queue  = pthread->type;

	queue_head_t *phdr = &threads_runq[queue];

	queue_remove(phdr, pthread, pthread_thread_t *, runq);
	pthread->runq.next = (queue_entry_t) 0;	

	queue_count[queue]--;

	threads_runq_count--;
	if (queue_empty(phdr))
		threads_whichrunq &= ~(1 << queue);
}

/*
 * The interface for the main scheduler
 */ 

void
methos_sched_init(void)
{
	int i;
	for (i = 0; i < NUMQUEUES; i++)
		queue_init(&threads_runq[i]);
}

/*
 * This scheduler has an entirely different understanding of policy
 */
int
methos_sched_schedules(int policy)
{
	return 1;
}

/*
 * Make a thread runnable. Called from the scheduler module.
 */
int
methos_sched_setrunnable(pthread_thread_t *pthread)
{
	int resched = 0;

	if (methos_runq_onrunq(pthread))
		panic("methos_sched_setrunnable: Already on runQ: 0x%x(%d)",
		      (int) pthread, pthread->tid);

	methos_runq_insert_tail(pthread);

	return resched;
}

/*
 * Terminate our association with a thread,
 */
void
methos_sched_disassociate(pthread_thread_t *pthread)
{
	pthread_lock(&pthread_sched_lock);
	if (methos_runq_onrunq(pthread)) {
		methos_runq_remove(pthread);
	}
	pthread_unlock(&pthread_sched_lock);
}

int
methos_sched_check_schedstate(const struct sched_param *param)
{
	return 0;
}

/*
 * Initialize scheduling parameters for a newly created thread. Note,
 * we don't need any locking since there are no side effects.
 */
void
methos_sched_init_schedstate(pthread_thread_t *pthread,
			    const struct sched_param *param)
{
	pthread->ticks		= INITIAL_TIMESLICE;
	pthread->timeslice	= INITIAL_TIMESLICE;
	pthread->type		= param->type;
}

/*
 * Change thread scheduler state. pthread is locked. interrupts are disabled.
 */
int
methos_sched_change_state(pthread_thread_t *pthread,
			 const struct sched_param *param)
{
	int newtype = param->type;
	int resched = 0;
	
	if (pthread->type == newtype)
	    goto done;

	pthread_lock(&pthread_sched_lock);

	if (methos_runq_onrunq(pthread)) {
		methos_runq_remove(pthread);
 		pthread->type = newtype;
		methos_runq_insert_tail(pthread);
	}
	else {
		pthread->type = newtype;
	}
	pthread_unlock(&pthread_sched_lock);
 done:
	return resched;
}

/*
 * Not necessary for this scheduler
 */
int
methos_sched_priority_bump(pthread_thread_t *pthread, int newprio)
{
	return 0;
}

/*
 * Dispatch a thread back to the runq.
 */
int
methos_sched_dispatch(resched_flags_t reason, pthread_thread_t *pthread)
{
	pthread->cputime++;
	tickcount++;
	if (tickcount > 100) {
		methos_sched_calc_percentages ();
		tickcount = 0;
	}

	switch (reason) {
		case RESCHED_USERYIELD:
			if (pthread->type == TTYPE_BACKGROUND) {
				if (pthread->timeslice > MIN_TIMESLICE) {
					pthread->timeslice--;
				}
			}

			methos_runq_insert_tail(pthread);
			break;
		
		case RESCHED_YIELD:
			if (pthread == IDLETHREAD)
				panic("methos_sched_dispatch: Idlethread!\n");

			if (pthread->type == TTYPE_BACKGROUND) {
				if (pthread->timeslice > MIN_TIMESLICE) {
					pthread->timeslice--;
				}
			}

			methos_runq_insert_head(pthread);
			break;
		
		case RESCHED_PREEMPT:
			if (pthread == IDLETHREAD)
				panic("methos_sched_dispatch: Idlethread!\n");

			if (--pthread->ticks == 0) {
				if (pthread->type == TTYPE_BACKGROUND) {
					if (pthread->timeslice < MAX_TIMESLICE) {
						pthread->timeslice++;
					}
				}
				pthread->ticks = pthread->timeslice;
				methos_runq_insert_tail(pthread);
			}
			else {
				if (pthread != idle_thread) {
					methos_runq_insert_head(pthread);
				}
			}
			
			break;

		case RESCHED_INTERNAL:
			methos_runq_insert_head (idle_thread);			
			break;
		
		default:
			break;
	}
	return 0;
}

/*
 * Return the next thread ready to run.
 */
pthread_thread_t *
methos_sched_thread_next(void)
{
	if (methos_runq_empty()) {
		return 0;
	}

	return methos_runq_dequeue();
}
